// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
/***/
/*
Converts a 1D buffer of Z values into a 2D grid given the input pointspacing. 
It assigns a Z value to the closest grid point in the X-Y plane. 
If no point falls in a given grid cell, valids flag is turn to 0 for that cell value.
Outputs:
- Z2: z values in a 2D grid (X-Y plane)
- valids: buffer with the validity flag equal to 1 to indicate that a point withing a pointspacing distance has been assigned, 0 otherwise
- idcorresp: buffer with the ID correspondences between 1D buffer and 2D grid points.
Inputs:
- x1: 2D buffer containing all the grid values in the X-axis
- y1: 2D buffer containing all the grid values in the Y-axis
- x2: 1D buffer containing the X-coordinate of the point cloud/mesh vertices
- y2: 1D buffer containing the Y-coordinate of the point cloud/mesh vertices
- z2: 1D buffer containing the Z-coordinate of the point cloud/mesh vertices that are to be assigned to a 2D grid
- radoffset: search radius in the 2-D grid to find potential points within pointspacing distance. (This is to avoid searching over millions of points everytime).
- pointspacing: this is the distance separation between cells in the 2D grid.

Version history: 2017-03-28	Marc Bosch	Initial version
*/
/***/
#include <math.h>
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits>
//#include <omp.h>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    if(nlhs > 3){
        mexErrMsgTxt("Function has three return value");
        return;
    }

    if(nrhs !=7){
        mexErrMsgTxt("Usage: [Z2,valids,idcorresp]=mexbuffer2gridC(x1,y1,x2,y2,z2,radoffset,pointspacing);\nmexbuffer2gridC(double *X2, double *Y2, double *bufferx, double *buffery, double *bufferz, double radoffset,double pointspacing)");
        return;
    }
    
    /* get the image buffer*/    
    double *X2 = (double*) mxGetPr(prhs[0]);
	double *Y2 = (double*) mxGetPr(prhs[1]);
    int* dims = (int*) mxGetDimensions(prhs[0]);
    int rows = dims[1];//flipped since it comes transpose and row major
    int cols = dims[0];
    double *bufferx = (double*) mxGetPr(prhs[2]);
	double *buffery = (double*) mxGetPr(prhs[3]);
	double *bufferz = (double*) mxGetPr(prhs[4]);
	double* pradoffset = mxGetPr(prhs[5]);
 	double radoffset = *pradoffset;
	double* ppointspacing = mxGetPr(prhs[6]);
 	double pointspacing = *ppointspacing;
	int* dimsbuf = (int*) mxGetDimensions(prhs[2]);
 	int len = MAX(dimsbuf[0],dimsbuf[1]);
    plhs[0] = mxCreateDoubleMatrix(rows,cols, mxREAL);
    double *Z2 = mxGetPr(plhs[0]);  
    for (int i=0; i < rows*cols; i++)
        Z2[i] = 0;//dims[0];
	plhs[1] = mxCreateDoubleMatrix(rows,cols, mxREAL);
    double *valids = mxGetPr(plhs[1]);  
    for (int i=0; i < rows*cols; i++)
        valids[i] = 0;//dims[0];  
	plhs[2] = mxCreateDoubleMatrix(rows,cols, mxREAL);
	double *idcorresp = mxGetPr(plhs[2]);  
	for (int i=0; i < rows*cols; i++)
        idcorresp[i] = 0;//dims[0];  
	int *previous = (int*)malloc(len*sizeof(int));
	for (int i=0; i < len; i++)
        previous[i] = 0;//dims[0];  
	int kbest=-1;
	int searchoffset=radoffset*cols;
	int kstart=0;
	int kend=len;
	int kstart_last = 0;
   //look for closer geometric point and assign the corresponding z value 
	int k=0;
	
	double dminMAX = std::numeric_limits<double>::max();
	const double pointspa = pointspacing*pointspacing*2;
	//#pragma omp parallel for
	for (int i=0;i<rows*cols;i++)
	{
		//printf("Num threads %d, thread ID %d.\n", omp_get_num_threads(), omp_get_thread_num());
		double dmin = dminMAX;
		int quicksearch = 0;
		//quick search		
		for (k=MAX(0,kbest-10);k<MIN(len,kbest+10);k++)
		{
			double dist=((X2[i]-bufferx[k])*(X2[i]-bufferx[k])+
			(Y2[i]-buffery[k])*(Y2[i]-buffery[k]));
			if (dist<dmin)
			{	
				dmin = dist;
				Z2[i] = bufferz[k];
				//kbest=k;
				idcorresp[i]=k;
				
				if (dmin<pointspa)//0.2)
				{
					kbest=k;
					previous[k]+=1;
					if(previous[k]==1)
					{
						quicksearch=1;	
						valids[i]=1;
					}
				}
			}		
		}
		//extended search
		if(quicksearch==0)
		{
			dmin = dminMAX;
			for (k=kstart;k<kend;k++)
			{
				double dist=((X2[i]-bufferx[k])*(X2[i]-bufferx[k])+
				(Y2[i]-buffery[k])*(Y2[i]-buffery[k]));
				if (dist<dmin)
				{	
					dmin = dist;
					Z2[i] = bufferz[k];
					//kbest=k;
					idcorresp[i]=k;
				
					if (dmin<pointspa)
					{
						kbest=k;
						previous[k]+=1;						
						if(previous[k]==1)
							break;
					}
				}
		
			}
			if (dmin<pointspa)// && previous[kbest]<2)
			{
				searchoffset=radoffset*cols;
				valids[i]=1;
				if (previous[k]>1)
					valids[i]=2;
			}
			else
			{
				searchoffset=2*radoffset*cols;
			}
		}
		
		
		kstart=MAX(0,kbest-searchoffset);
		kend=MIN(len,kbest+searchoffset);		
		if(i%(int)cols==0)
			kstart_last = kbest;		
		//re-set for each new column/row
		if(i%(int)cols==(int)cols-1)
		{			
			kstart=MAX(0,kstart_last-searchoffset);
			kend=MIN(len,kstart_last+searchoffset);	
		}
		
	}
	
   free(previous);
}
